/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "html/align.hh"
#include "html/boolAttr.hh"
#include "html/genAttr.hh"
#include "html/size.hh"
#include "html/table.hh"
#include "html/text.hh"
#include "html/writer.hh"
#include <stdexcept>
#include "PConfig.h"
#include <sstream>

using namespace std;
using namespace html;

//======================================  Construct an empty cell
cell::cell(void)
   : mColSpan(1)
{}

//======================================  Destroy a cell
cell::~cell(void) {}

//======================================  Clone a cell
cell*
cell::clone(void) const {
   return new cell(*this);
}

//======================================  Set column span
void
cell::colspan(int span) {
   mColSpan = span;
   if (span > 1) {
      addAttr("colspan", genAttr(to_string(span)));
   }
}

//======================================  Set the alignment attribute
void
cell::setAlign(const std::string& a) {
   addAttr("align", align(a));
}

//======================================  set the Save attribute
void
cell::setSave(bool a) {
   addAttr("save", boolAttr(a));
}

//======================================  Header cell constructor
header_cell::header_cell(const std::string& title) {
   if (!title.empty()) addObject(text(title));
}

//======================================  Header cell destructor.
header_cell::~header_cell(void) {
}

//======================================  Clone a header cell
header_cell* 
header_cell::clone(void) const {
   return new header_cell(*this);
}

//======================================  Construct an empty table
table::table(void)
   : mRow(0)
{}

//======================================  Construct an empty table
table::table(const string& title) 
   : mTitle(title), mRow(0)
{}

//======================================  Destroy a table
table::~table(void) {
}

//======================================  Clone a table
table* 
table::clone(void) const {
   return new table(*this);
}

//======================================  Write out the html
void 
table::write(writer& out) const {
   out.tag("blockquote");
   out.tag("center");
   if (!mTitle.empty()) {
      html::text(mTitle).setSize(size(1)).write(out);
      out.endLine();
   }
   out.tag("table", mAttrib);
   out.endLine();

   //----------------------------------  Print out the column headers
   int nCol = getNColumn();
   out.tag("thead");
   out.tag("tr");
   for (int i=0 ; i<nCol ; i++) {
      if (mCTitle[i].colspan()) mCTitle[i].write(out);
   }
   out.endTag("tr");
   out.endTag("thead");
   out.endLine();

   //----------------------------------  Print the column data
   out.tag("tbody");
   for (int j=0 ; j<mRow ; j++) {
      out.tag("tr");
      int row0 = j*nCol;
      for (int i=0 ; i<nCol ; i++) {
	 if (mData[i+row0]) {
	    mData[i+row0]->write(out);
	 } else {
	    out.tag("td");
	    out.endTag("td");
	 }
      }
      out.endTag("tr");
      out.endLine();
   }
   out.endTag("tbody");
   out.endLine();

   out.endTag("table");
   out.endTag("center");
   out.endTag("blockquote");
}

//======================================  Add a column
int
table::addColumn(const string& title) {
   mCTitle.push_back(header_cell(title));
   return mCTitle.size()-1;
}

//======================================  Add a row
int
table::addRow(void) {
   mData.resize((mRow+1)*getNColumn());
   return mRow++;
}

//======================================  Add a row
int
table::addSubHeaderRow(void) {
   header_cell hc;
   object_ptr op(hc);
   mData.resize((mRow+1)*getNColumn(), op);
   return mRow++;
}

//======================================  Erase one or more rows
void 
table::eraseRow(int row, int count)
{
   if (row < 0 || row >= mRow) throw runtime_error("Invalid row number");
   int nCol = getNColumn();
   int maxRow = row + count;
   if (maxRow > mRow) maxRow = mRow;
   mData.erase(mData.begin()+row*nCol, mData.begin()+maxRow*nCol);
   mRow -= maxRow - row;
}


//======================================  Insert a number into a cell
void 
table::insertData(int row, int col, double x, int wid, int prec) 
  
{
   ostringstream ss;
   if (wid>0)  ss.width(wid);
   if (prec>=0) {
      ss << fixed;
      ss.precision(prec);
   }
   ss << x;
#ifdef __GNU_STDC_OLD
   ss << ends;
#endif
   insertData(row, col, text(ss.str()));
   refCell(row, col).setAlign("right");
}

//======================================  Insert a number into a cell
void 
table::insertData(int row, int col, long x, int wid, int radix) 
  
{
   ostringstream ss;
   if (wid>0)  ss.width(wid);
   switch (radix) {
   case 10:
      break;
   case 16:
      ss << hex;
      break;
   case 8:
      ss << dec;
      break;
   default:
      throw runtime_error("Invalid radix");
   }
   ss << x;
#ifdef __GNU_STDC_OLD
   ss << ends;
#endif
   insertData(row, col, text(ss.str()));
   refCell(row, col).setAlign("right");
}

//======================================  Insert text into a cell
void 
table::insertData(int row, int col, const string& txt) 
{
   insertData(row, col, text(txt));
}

//======================================  Insert data into a cell
void 
table::insertData(int row, int col, const object& obj) 
{
   refCell(row, col).addObject(obj);
}

//======================================  Insert a row in front of "row"
void 
table::insertRow(int row)
{
   mData.insert(mData.begin()+inxrc(row, 0), getNColumn(), object_ptr());
}

//======================================  Reference a constant cell
const html::cell& 
table::refCell(int row, int col) const {
   const static cell empty_cell;
   int inx = inxrc(row, col);
   if (!mData[inx]) return empty_cell;
   return dynamic_cast<const cell&>(*mData[inx]);
}

//======================================  Reference a cell
html::cell& 
html::table::refCell(int row, int col) {
   int inx = inxrc(row, col);
   if (!mData[inx]) mData[inx] = new cell;
   return dynamic_cast<cell&>(*mData[inx]);
}

//======================================  Reference a constant header cell
const html::cell& 
html::table::refHeader(int col) const {
   if (col<0 || col>=getNColumn()) throw std::runtime_error("Invalid Column.");
   return mCTitle[col];
}

//======================================  Reference a Header cell
html::cell& 
html::table::refHeader(int col) {
   if (col<0 || col>=getNColumn()) throw std::runtime_error("Invalid Column.");
   return mCTitle[col];
}

//======================================  Add/Remove the border attribute
void
table::setBorder(bool On) {
   if (On) mAttrib.addAttr("border", boolAttr());
   else    mAttrib.remAttr("border");
}

//======================================  Add/Modify the class attribute
void
table::setClass(const std::string& class_name) {
   mAttrib.appendAttr("class", genAttr(class_name), ',');
}

//======================================  Span cells
void
table::span(int row, int col, int nspan) {
   if (col < 0 || nspan < 1 || col+nspan > mCTitle.size()) {
      throw runtime_error("html::table: Invalid span");
   }
   //-----------------------------------  span header columns
   if (row < 0) {
      refHeader(col).colspan(nspan);
      for (int icol=col+1; icol<col+nspan; icol++) {
	 refHeader(icol).colspan(0);
      }
   }
   //-----------------------------------  span table columns
   else {
      refCell(row, col).colspan(nspan);
      for (int icol=col+1; icol<col+nspan; icol++) {
	 refCell(row, icol).colspan(0);
      }
   }
}
